from typing import List

from opentelemetry import trace
from opentelemetry.trace import Status, StatusCode
from opentelemetry.sdk.trace import TracerProvider
from starlette.requests import Request
from starlette.routing import Match
from starlette.status import HTTP_500_INTERNAL_SERVER_ERROR
from starlette.types import Message, Receive, Scope, Send
from . import monitoring_metrics as metrics


class MonitoringMiddleware:
    def __init__(
        self,
        app,
        app_name: str = "fastapi-app",
        tracer: TracerProvider = None,
        exclude_paths: List = None,
    ):
        self.app = app
        self.app_name = app_name
        self.tracer = trace.get_tracer("cdp-monitoring", tracer_provider=tracer)
        self.exclude_paths = exclude_paths or []

        metrics.INFO.labels(app_name=self.app_name).inc()

    @staticmethod
    def _get_endpoint(request: Request):
        for route in request.app.routes:
            match, child_scope = route.matches(request.scope)
            if match == Match.FULL:
                return True, route.endpoint

        return False, None

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        if scope["path"] in self.exclude_paths:
            return await self.app(scope, receive, send)

        request = Request(scope, receive=receive)

        is_handled_path, endpoint = self._get_endpoint(request)
        if not is_handled_path:
            return await self.app(scope, receive, send)

        async def receive_wrapper() -> Message:
            parent_span = trace.get_current_span()

            with self.tracer.start_as_current_span("http.request") as span:
                message = await receive()

                if message["type"] == "http.request":
                    request_body = message["body"].decode("utf-8")
                    parent_span.set_attribute("request.body", request_body)
                    span.set_attribute("request.body", request_body)

                return message

        async def send_wrapper(message: Message) -> None:
            parent_span = trace.get_current_span()

            if message["type"] == "http.response.start":
                parent_span.set_attribute("response.status_code", message["status"])
                parent_span.set_attribute("response.headers", str(message["headers"]))

                with self.tracer.start_as_current_span("http.response.start") as span:
                    span.set_attribute("response.status_code", message["status"])
                    span.set_attribute("response.headers", str(message["headers"]))

                    await send(message)

                    metrics.RESPONSES.labels(
                        method=scope["method"],
                        path=scope["path"],
                        status_code=message["status"],
                        app_name=self.app_name,
                    ).inc()

            if message["type"] == "http.response.body":
                parent_span.set_attribute("response.body", message["body"])

                with self.tracer.start_as_current_span("http.response.body") as span:
                    span.set_attribute("response.body", message["body"])

                    await send(message)

        metrics.REQUESTS.labels(
            method=scope["method"], path=scope["path"], app_name=self.app_name
        ).inc()

        with self.tracer.start_as_current_span(
            f"{scope['method']} {scope['path']}"
        ) as span:
            try:
                trace_id = trace.format_trace_id(span.get_span_context().trace_id)

                span.set_attribute("protocol", str(scope["type"]))
                span.set_attribute("http_version", str(scope["http_version"]))
                span.set_attribute("server.host", str(scope["server"][0]))
                span.set_attribute("server.port", str(scope["server"][1]))
                span.set_attribute("client.host", str(scope["client"][0]))
                span.set_attribute("client.port", str(scope["client"][1]))
                span.set_attribute("request.headers", str(scope["headers"]))
                span.set_attribute("request.query_string", str(scope["query_string"]))

                with metrics.REQUESTS_PROCESSING_TIME.labels(
                    method=scope["method"],
                    path=scope["path"],
                    app_name=self.app_name,
                ).time(exemplar={"TraceID": trace_id}):
                    with self.tracer.start_as_current_span(endpoint.__name__) as span:
                        await self.app(scope, receive_wrapper, send_wrapper)

            except Exception as e:
                metrics.EXCEPTIONS.labels(
                    method=scope["method"],
                    path=scope["path"],
                    exception_type=type(e).__name__,
                    app_name=self.app_name,
                ).inc()

                metrics.RESPONSES.labels(
                    method=scope["method"],
                    path=scope["path"],
                    status_code=HTTP_500_INTERNAL_SERVER_ERROR,
                    app_name=self.app_name,
                ).inc()

                span.set_status(Status(StatusCode.ERROR))
                span.record_exception(e)
                raise e from None



from timeit import default_timer
from prometheus_client import Counter, Gauge
from prometheus_client.context_managers import Timer as BaseTimer
from prometheus_client.metrics import Histogram as BaseHistogram


class Timer(BaseTimer):
    def __init__(self, metric, callback_name, *args, **kwargs):
        super().__init__(metric, callback_name)
        self._callback_args = args
        self._callback_kwargs = kwargs

    def __exit__(self, typ, value, traceback):
        duration = max(default_timer() - self._start, 0)
        callback = getattr(self._metric, self._callback_name)
        callback(duration, *self._callback_args, **self._callback_kwargs)


class Histogram(BaseHistogram):
    def time(self, *args, **kwargs):
        return Timer(self, "observe", *args, **kwargs)


INFO = Gauge("fastapi_app_info", "FastAPI application information.", ["app_name"])
REQUESTS = Counter(
    "fastapi_requests_total",
    "Total count of requests by method and path.",
    ["method", "path", "app_name"],
)
RESPONSES = Counter(
    "fastapi_responses_total",
    "Total count of responses by method, path and status codes.",
    ["method", "path", "status_code", "app_name"],
)
REQUESTS_PROCESSING_TIME = Histogram(
    "fastapi_requests_duration_seconds",
    "Histogram of requests processing time by path (in seconds)",
    ["method", "path", "app_name"],
)
EXCEPTIONS = Counter(
    "fastapi_exceptions_total",
    "Total count of exceptions raised by path and exception type",
    ["method", "path", "exception_type", "app_name"],
)


__all__ = [
    "INFO",
    "REQUESTS",
    "RESPONSES",
    "REQUESTS_PROCESSING_TIME",
    "EXCEPTIONS",
]



import os
import logging
from typing import List
from datetime import datetime

from fastapi import FastAPI
from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from prometheus_client import REGISTRY, CollectorRegistry
from prometheus_client.multiprocess import MultiProcessCollector
from prometheus_client.openmetrics.exposition import (
    CONTENT_TYPE_LATEST,
    generate_latest,
)
from starlette.responses import Response
from .monitoring_middleware import MonitoringMiddleware

logger = logging.getLogger(__name__)


def metrics(request):
    # multiprocessing 환경일 때 필수
    if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
        registry = CollectorRegistry()
        MultiProcessCollector(registry)

        return Response(
            generate_latest(registry), headers={"Content-Type": CONTENT_TYPE_LATEST}
        )

    return Response(
        generate_latest(REGISTRY), headers={"Content-Type": CONTENT_TYPE_LATEST}
    )


def init_monitoring(app: FastAPI, version: str = None, exclude_paths: List[str] = None):
    """
    fast api app 에 프로메테우스 모니터링 및 오픈텔레메트리 트레이싱을 적용
    :param app:
    :param version: application 의 버전 (없으면 현재 시간을 prometheus app_name 으로 사용)
    :param exclude_paths: 모니터링을 적용하지 않을 path
    :return:
    """
    endpoint = os.environ.get("TEMPO_URL", None)
    if endpoint is None:
        logger.info("TEMPO_URL is not set. Not available now")
        return None

    if version is None:
        # version 이 주어지지 않으면 앱 실행 시간을 postfix 로 사용
        now = str(datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        app_name = f'{app.title}-{now}'
    else:
        app_name = f'{app.title}-{version}'

    resource = Resource.create(
        attributes={"service.name": app_name, "compose_service": app_name}
    )

    tracer = TracerProvider(resource=resource)
    trace.set_tracer_provider(tracer)
    tracer.add_span_processor(BatchSpanProcessor(OTLPSpanExporter(endpoint=endpoint)))

    app.add_middleware(
        MonitoringMiddleware,
        app_name=app_name,
        tracer=tracer,
        exclude_paths=exclude_paths,
    )
    app.add_route("/metrics", metrics)
